#ifndef __SCANNER_H__
#define __SCANNER_H__

#include "parsetools.h"

//////////////////////////////////////////////////////////////////////////////
// CFileFrame Class
//

class CFileFrame
{
public:
  CFileFrame();
  ~CFileFrame();
  VOID Init(CHAR *pcName, FILE *fpFile, YY_BUFFER_STATE yyBufferState,
	    UINT_32 uiLine, UINT_32 uiColumn);
  VOID Uninit();

public:
  CHAR *m_pcName;
  FILE *m_fpFile;
  YY_BUFFER_STATE m_yyBufferState;
  // These Variables come from the previous frame
  UINT_32 m_uiLine;
  UINT_32 m_uiColumn;
};

//////////////////////////////////////////////////////////////////////////////
// CScanner Class
//

typedef VOID (YY_BEGIN_INITIAL_TYPE) ();
typedef YY_BUFFER_STATE (YY_GET_CURRENT_BUFFER_TYPE) ();
typedef YY_BUFFER_STATE (YY_CREATE_BUFFER_TYPE) (FILE*);
typedef VOID (YY_DELETE_BUFFER_TYPE) (YY_BUFFER_STATE);
typedef VOID (YY_SWITCH_TO_BUFFER_TYPE) (YY_BUFFER_STATE);
typedef VOID (YY_SET_VALUE_INT_TYPE) (int);
typedef VOID (YY_SET_VALUE_FLOAT_TYPE) (float);
typedef VOID (YY_SET_VALUE_STRING_TYPE) (CHAR*);

class CScanner
{
public:
  CScanner();
  ~CScanner();

  VOID Init(CHAR *pcFile, CHAR **ppcDirectories, UINT_32 uiDirectories, 
	    YYLTYPE *pyylloc, int *pyyleng,
	    YY_BEGIN_INITIAL_TYPE *pfyy_BeginInitial,
	    YY_GET_CURRENT_BUFFER_TYPE *pfyy_GetCurrentBuffer,
	    YY_CREATE_BUFFER_TYPE *pfyy_CreateBuffer,
	    YY_DELETE_BUFFER_TYPE *pfyy_DeleteBuffer,
	    YY_SWITCH_TO_BUFFER_TYPE *pfyy_SwitchToBuffer,
	    YY_SET_VALUE_INT_TYPE *pfyy_SetValueInt,
	    YY_SET_VALUE_FLOAT_TYPE *pfyy_SetValueFloat,
	    YY_SET_VALUE_STRING_TYPE *pfyy_SetValueString);
  VOID Uninit();

public:
  VOID ScanInteger(CHAR *pcToken);
  VOID ScanFloat(CHAR *pcToken);
  VOID ScanString();
  VOID ScanTclBlock();
  CHAR *ScanFileName();
  VOID ScanIdentifier(CHAR *pcToken);

  VOID PushFile(CHAR *pcFile);
  BOOL PopFile();

  VOID HandleError(UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcFormat, ...);
  VOID HandleWarning(UINT_32 uiLine, UINT_32 uiColumn, CHAR *pcFormat, ...);

public:
  inline UINT_32 GetErrorCount();
  inline UINT_32 GetWarningCount();

  inline VOID ConsumeToken();
  inline VOID ConsumeNewLine();

  inline VOID CommentStart();
  inline BOOL CommentEnd();

  inline VOID StringStart();
  inline VOID StringEnd();
  inline VOID StringAppendNewLine();
  inline VOID StringAppend(CHAR *pcBuffer);

  inline VOID TclBlockStart();
  inline BOOL TclBlockEnd();
  inline VOID TclBlockAppendNewLine();
  inline VOID TclBlockAppend(CHAR *pcBuffer);

  inline VOID FileNameStart();
  inline VOID FileNameEnd();
  inline VOID FileNameAppendNewLine();
  inline VOID FileNameAppend(CHAR *pcBuffer);

protected:
  FILE *m_fpError;

  CHashString m_hashIdentifierTable;
  CHashString m_hashStringTable;
  CHashString m_hashFileNameTable;
  CHashString m_hashOpenFileTable;
  CListInt m_listOpenFileStack;

  UINT_32 m_uiErrorCount;
  UINT_32 m_uiWarningCount;

  //CHAR *m_pcCurrFile;
  //UINT_32 m_uiCurrLine;
  //UINT_32 m_uiCurrColumn;

  // Comments
  UINT_32 m_uiCommentLevel;
  UINT_32 m_uiCommentStartLine;
  UINT_32 m_uiCommentStartColumn;

  // Strings
  CHAR m_pcStringBuffer[200];
  CHAR *m_pcStringBufferPos;
  UINT_32 m_uiStringStartLine;
  UINT_32 m_uiStringStartColumn;

  // Tcl Blocks
  CHAR m_pcTclBlockBuffer[10000];
  CHAR *m_pcTclBlockBufferPos;
  UINT_32 m_uiTclBlockLevel;
  UINT_32 m_uiTclBlockStartLine;
  UINT_32 m_uiTclBlockStartColumn;

  // FileNames
  CHAR m_pcFileNameBuffer[200];
  CHAR *m_pcFileNameBufferPos;
  UINT_32 m_uiFileNameStartLine;
  UINT_32 m_uiFileNameStartColumn;

  // Path Search
  CHAR **m_ppcDirectories;
  UINT_32 m_uiDirectories;

  YYLTYPE *m_pyylloc;
  int *m_pyyleng;

  YY_BEGIN_INITIAL_TYPE *m_pfyy_BeginInitial;
  YY_GET_CURRENT_BUFFER_TYPE *m_pfyy_GetCurrentBuffer;
  YY_CREATE_BUFFER_TYPE *m_pfyy_CreateBuffer;
  YY_DELETE_BUFFER_TYPE *m_pfyy_DeleteBuffer;
  YY_SWITCH_TO_BUFFER_TYPE *m_pfyy_SwitchToBuffer;
  YY_SET_VALUE_INT_TYPE *m_pfyy_SetValueInt;
  YY_SET_VALUE_FLOAT_TYPE *m_pfyy_SetValueFloat;
  YY_SET_VALUE_STRING_TYPE *m_pfyy_SetValueString;
};

//////////////////////////////////////////////////
// Inlined Functions
//

inline UINT_32 CScanner::GetErrorCount()
{
  return m_uiErrorCount;
}

inline UINT_32 CScanner::GetWarningCount()
{
  return m_uiWarningCount;
}

inline VOID CScanner::ConsumeToken()
{
  m_pyylloc->first_column += *m_pyyleng;
}

inline VOID CScanner::ConsumeNewLine()
{
  m_pyylloc->first_line++;
  m_pyylloc->first_column = 1;
}

inline VOID CScanner::CommentStart()
{
  if ( m_uiCommentLevel == 0 )
    {
      m_uiCommentStartLine = m_pyylloc->first_line;
      m_uiCommentStartColumn = m_pyylloc->first_column;
    }

  m_uiCommentLevel++;
  ConsumeToken();
}

inline BOOL CScanner::CommentEnd()
{
  BOOL bDone;

  bDone = FALSE;
  m_uiCommentLevel--;
  if ( m_uiCommentLevel == 0 )
    {
      m_uiCommentStartLine = 0;
      m_uiCommentStartColumn = 0;
      bDone = TRUE;
    }

  ConsumeToken();

  return bDone;
}


inline VOID CScanner::StringStart()
{
  m_pcStringBufferPos = m_pcStringBuffer;
  m_uiStringStartLine = m_pyylloc->first_line;
  m_uiStringStartColumn = m_pyylloc->first_column;

  ConsumeToken();
}

inline VOID CScanner::StringEnd()
{
  m_pcStringBufferPos[0] = '\0';
  m_uiStringStartLine = 0;
  m_uiStringStartColumn = 0;

  ConsumeToken();
}

inline VOID CScanner::StringAppendNewLine()
{
  m_pcStringBufferPos[0] = '\n';
  m_pcStringBufferPos++;

  ConsumeNewLine();
}

inline VOID CScanner::StringAppend(CHAR *pcBuffer)
{
  ASSERT( pcBuffer != NULL );

  strcpy(m_pcStringBufferPos, pcBuffer);
  m_pcStringBufferPos += strlen(pcBuffer);

  ConsumeToken();
}

inline VOID CScanner::TclBlockStart()
{
  if ( m_uiTclBlockLevel == 0 )
    {
      m_pcTclBlockBufferPos = m_pcTclBlockBuffer;
      
      m_uiTclBlockStartLine = m_pyylloc->first_line;
      m_uiTclBlockStartColumn = m_pyylloc->first_column;
    }
  else
    {
      m_pcTclBlockBufferPos[0] = '{';
      m_pcTclBlockBufferPos++;
    }

  m_uiTclBlockLevel++;
  ConsumeToken();
}

inline BOOL CScanner::TclBlockEnd()
{
  BOOL bDone;

  bDone = FALSE;
  m_uiTclBlockLevel--;

  if ( m_uiTclBlockLevel == 0 )
    {
      m_pcTclBlockBufferPos[0] = '\0';

      m_uiTclBlockStartLine = 0;
      m_uiTclBlockStartColumn = 0;
      bDone = TRUE;
    }
  else
    {
      m_pcTclBlockBufferPos[0] = '}';
      m_pcTclBlockBufferPos++;
    }

  ConsumeToken();

  return bDone;
}

inline VOID CScanner::TclBlockAppendNewLine()
{
  m_pcTclBlockBufferPos[0] = '\n';
  m_pcTclBlockBufferPos++;

  ConsumeNewLine();
}

inline VOID CScanner::TclBlockAppend(CHAR *pcBuffer)
{
  ASSERT( pcBuffer != NULL );

  strcpy(m_pcTclBlockBufferPos, pcBuffer);
  m_pcTclBlockBufferPos += strlen(pcBuffer);

  ConsumeToken();
}

inline VOID CScanner::FileNameStart()
{
  m_pcFileNameBufferPos = m_pcFileNameBuffer;
  m_uiFileNameStartLine = m_pyylloc->first_line;
  m_uiFileNameStartColumn = m_pyylloc->first_column;

  ConsumeToken();
}

inline VOID CScanner::FileNameEnd()
{
  m_pcFileNameBufferPos[0] = '\0';
  m_uiFileNameStartLine = 0;
  m_uiFileNameStartColumn = 0;

  ConsumeToken();
}

inline VOID CScanner::FileNameAppendNewLine()
{
  m_pcFileNameBufferPos[0] = '\n';

  ConsumeNewLine();
}

inline VOID CScanner::FileNameAppend(CHAR *pcBuffer)
{
  ASSERT( pcBuffer != NULL );

  strcpy(m_pcFileNameBufferPos, pcBuffer);
  m_pcFileNameBufferPos += strlen(pcBuffer);

  ConsumeToken();
}

#endif // __SCANNER_H__


