#include "sliderender.h"

#include "slideworld.h"
#include "slideviewport.h"
#include "slidecamera.h"
#include "slidenode.h"
#include "slideface.h"
#include "slidesurface.h"

static CHAR gpcBufferVariable[200];
static CHAR gpcBufferValue[200];

BOOL CSLIDERender::m_bRenderClipBound = FALSE;
BOOL CSLIDERender::m_bCrystalBallEnabled = FALSE;

UINT_32 CSLIDERender::m_uiCount = 0;
Vector CSLIDERender::m_vProjMin;
Vector CSLIDERender::m_vProjMax;
CBound CSLIDERender::m_boundClip;

CListInt *CSLIDERender::m_plistCSLIDETclVariables = NULL;

//////////////////////////////////////////////////////////////////////
// CSLIDERender Class
//

CSLIDERender::CSLIDERender()
{
  if ( m_uiCount == 0 )
    {
      // These are the limits of the canonical view volume
      m_vProjMin.Set(-1, -1, -1);
      m_vProjMax.Set(1, 1, 0);

      // This a bound around the canonical view volume which
      // we will use to do clipping with homogeneous coordinates
      Point pt;
      
      pt = m_vProjMin;
      m_boundClip.Insert(pt);
      pt = m_vProjMax;
      m_boundClip.Insert(pt);

      m_plistCSLIDETclVariables = new CListInt;
      ASSERT( m_plistCSLIDETclVariables != NULL );
      m_plistCSLIDETclVariables->Init();
    }
  m_uiCount++;

  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = NULL;

  m_pCSLIDECameraPath = NULL;

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);
}

CSLIDERender::~CSLIDERender()
{
  CSLIDETclVariables *pCSLIDETclVariables;

  m_uiCount--;
  if ( m_uiCount == 0 )
    {
      if ( m_plistCSLIDETclVariables == NULL )
	{
	  for ( pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst();
		pCSLIDETclVariables != NULL;
		pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->RemoveFirst() )
	    {
	      delete pCSLIDETclVariables;
	    }
	  
	  delete m_plistCSLIDETclVariables;
	  m_plistCSLIDETclVariables = NULL;
	}
    }
}

VOID CSLIDERender::Init(CHAR *pcFile, UINT_32 uiLine, UINT_32 uiColumn, 
			CHAR *pcViewportID, CSLIDECameraPath *pCSLIDECameraPath,
			CHAR *pcWorldRootID)
{
  CSLIDEPrimitive::Init(pcFile, uiLine, uiColumn, NULL);

  ASSERT( pcViewportID != NULL );
  ASSERT( pCSLIDECameraPath != NULL );
  ASSERT( pcWorldRootID != NULL );

  m_pcTclTag = NULL;
  m_pCSLIDETclVariables = NULL;

  m_pcViewportID = pcViewportID;

  m_pCSLIDECameraPath = pCSLIDECameraPath;

  m_pcWorldRootID = pcWorldRootID;
  m_pWorldRoot = NULL;

  m_dyniLOD.SetValue(slidelodtSLF_FULL);

  // Crystal Ball Variables
  m_iButtonCurr = GLSI_INVALID_BUTTON;
  m_fScaleLast = 1.0;
}

VOID CSLIDERender::Uninit()
{
  CSLIDEPrimitive::Uninit();

  m_pcViewportID = NULL;

  if ( m_pCSLIDECameraPath != NULL )
    {
      delete m_pCSLIDECameraPath;
      m_pCSLIDECameraPath = NULL;
    }

  m_pcWorldRootID = NULL;
  m_pWorldRoot = NULL;
}

BOOL CSLIDERender::Initialized()
{
  if ( !CSLIDEPrimitive::Initialized() )
    {
      return FALSE;
    }

  return ( m_pcViewportID != NULL &&
	   m_pCSLIDECameraPath != NULL &&
           m_pcWorldRootID != NULL );
}

BOOL CSLIDERender::CheckSemantics(CSLIDEWorld *pCSLIDEWorld)
{
  BOOL bCorrect, b;
  CSLIDEViewport *pCSLIDEViewport;

  if ( !CSLIDEPrimitive::CheckSemantics(pCSLIDEWorld) )
    {
      return FALSE;
    }

  bCorrect = TRUE;

  // Check Tcl Tag to see if we are sharing UI Variables

  m_pCSLIDETclVariables = (CSLIDETclVariables*)m_plistCSLIDETclVariables->Peek((INT_32)m_pcTclTag);
  if ( m_pCSLIDETclVariables == NULL )
    {
      m_pCSLIDETclVariables = new CSLIDETclVariables;
      ASSERT( m_pCSLIDETclVariables != NULL);
      m_pCSLIDETclVariables->Init(m_pcTclTag);

      m_plistCSLIDETclVariables->Insert(m_pCSLIDETclVariables, 
					(INT_32)m_pcTclTag);
    }

  // Check Viewport Reference

  pCSLIDEViewport = pCSLIDEWorld->PeekViewport(m_pcViewportID);
  if ( pCSLIDEViewport == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtViewport],
		  m_pcViewportID);
      bCorrect = FALSE;
    }
  else
    {
      pCSLIDEViewport->InsertRender(this);
    }

  // Check Camera Reference

  b = m_pCSLIDECameraPath->CheckSemantics(pCSLIDEWorld);
  bCorrect = bCorrect && b;

  // Check Root Node Reference

  m_pWorldRoot = pCSLIDEWorld->PeekNode(m_pcWorldRootID);
  if ( m_pWorldRoot == NULL )
    {
      HandleError("Undefined %s \"%s\"", 
		  gcppcSLIDEReservedWords[sliderwtObject],
		  m_pcWorldRootID);
      bCorrect = FALSE;
    }

  return bCorrect;
}

BOOL CSLIDERender::Valid()
{
  if ( !CSLIDEPrimitive::Valid() )
    {
      return FALSE;
    }

  return (m_pWorldRoot != NULL &&
	  ( 0 <= m_dyniLOD.GetValue() &&
	    m_dyniLOD.GetValue() < slidelodtSize ) );
}

VOID CSLIDERender::PreprocessGraph()
{
  m_pWorldRoot->PreprocessGraph();
}

VOID CSLIDERender::UpdateGraph()
{
  m_pWorldRoot->UpdateGraph();
}

VOID CSLIDERender::CheckDynamic()
{
  BOOL bDynamic;

  bDynamic = m_dyniLOD.GetDynamic();

  m_pCSLIDECameraPath->Preprocess();
  bDynamic = bDynamic || m_pCSLIDECameraPath->GetDynamic();

  SetDynamic(bDynamic);
}

VOID CSLIDERender::UpdateState()
{
  m_dyniLOD.Update();
  if ( !( 0 <= m_dyniLOD.GetValue() &&
	  m_dyniLOD.GetValue() < slidelodtSize ) )
    {
      HandleError("Illegal LOD flag value:  %ld", 
		  m_dyniLOD.GetValue());
      m_dyniLOD.SetValue(slidelodtSLF_FULL);
    }

  if ( m_dyniLOD.GetValue() != slidelodtSLF_OFF )
    {
      m_pCSLIDECameraPath->Update();
    }
}

VOID CSLIDERender::HandleReshape(const Vector &vSize)
{
  if ( m_dyniLOD.GetValue() != slideoftSLF_OFF )
    {
      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
    }
}

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: Crystal Ball UI
// 

// Both Mouse Functions get the size of the viewport rectangle
// and the location of the mouse cursor relative to the lower 
// left hand corner
BOOL CSLIDERender::HandleMouseButton(const Vector &vSize,
				     INT_32 iButton, INT_32 iState, 
				     const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( iState == GLSI_DOWN )
    {
      if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
	{
	  return FALSE;
	}

      m_vStart = vClick - (vSize / 2.0);
      m_vStart *= (FLOAT)(2.0 / MIN(vSize(X), vSize(Y)));
      m_vLatest = m_vStart;
      m_iButtonCurr = iButton;

      switch ( iButton ) 
	{
	case GLSI_LEFT_BUTTON:
	  break;
	case GLSI_MIDDLE_BUTTON:
	  m_fScaleLast = 1.0;
	  break;
	case GLSI_RIGHT_BUTTON:
	  break;
	default:
	  ASSERT(FALSE);
	  break;
	}
    }
  else if ( iState == GLSI_UP && iButton == m_iButtonCurr )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
    }

  return bPostRedisplay;
}

// We change the vClick into a vector from the center of the window
// where the radius of the best fit sphere is 1.0
BOOL CSLIDERender::HandleMouseMotion(const Vector &vSize,
				     const Vector &vClick)
{
  BOOL bPostRedisplay;

  bPostRedisplay = FALSE;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      m_iButtonCurr = GLSI_INVALID_BUTTON;
      return FALSE;
    }

  if ( m_iButtonCurr != GLSI_INVALID_BUTTON )
    {
      Matrix m, mInverse;
      Vector vCurr, v1, v2, vNorm;
      FLOAT theta;

      m_pCSLIDETclVariables->SetWidth(vSize(X));
      m_pCSLIDETclVariables->SetHeight(vSize(Y));
      m_pCSLIDETclVariables->SetMouseX(vClick(X)/vSize(X));
      m_pCSLIDETclVariables->SetMouseY(vClick(Y)/vSize(Y));
      
      vCurr = vClick - (vSize / 2.0);
      vCurr *= (FLOAT)(2.0 / MIN(vSize(X), vSize(Y)));

      m.SetIdentity();
      mInverse.SetIdentity();

      switch ( m_iButtonCurr ) 
	{
	case GLSI_LEFT_BUTTON:
	  {
	    // General Rotations

	    bPostRedisplay = TRUE;
	  }
	  break;
	case GLSI_MIDDLE_BUTTON:
	  {
	    // Scaling should change exponentially

	    bPostRedisplay = TRUE;
	  }
	  break;
	case GLSI_RIGHT_BUTTON:
	  {
	    // Rotations around the Z-axis

	    bPostRedisplay = TRUE;
	  }
	  break;
	}

      // Update the Matrix if it has changed
      if ( bPostRedisplay )
	{
	  Matrix mGvrp_vrc, mGvrc_vrp;

	  mGvrp_vrc = m_pCSLIDECameraPath->GetCamera()->GetGvrp_vrc();
	  mGvrc_vrp = m_pCSLIDECameraPath->GetCamera()->GetGvrc_vrp();

	  // Bring the Center of the VRC to the origin before applying 
          // scales and rotates
	  m_mBall = m_pCSLIDECameraPath->GetGroot_obj() * mGvrc_vrp * m * 
	            mGvrp_vrc * m_pCSLIDECameraPath->GetGobj_root() * m_mBall;
	  m_mBallInverse = m_mBallInverse * 
	                   m_pCSLIDECameraPath->GetGroot_obj() * mGvrc_vrp * 
	                   mInverse * 
	                   mGvrp_vrc * m_pCSLIDECameraPath->GetGobj_root();

	  // Even if the world is paused we want the UI to take effect
	  CSLIDEWorld::GetWorld()->ScheduleDynamics(0);
	}

      m_vLatest = vCurr;
    }

  return bPostRedisplay;
}
// 
// END  : ASSIGNMENT 5: Crystal Ball UI
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: Deal with the camera
// 
VOID CSLIDERender::Render(const Vector &vVPOrigin, const Vector &vVPSize)
{
  Matrix mGvrc_obj, mGobj_vrc;
  Matrix mGproj_vrc, mGvrc_proj;
  Matrix mGvp_proj;
  Vector vFrustum, vVPMax;

  if ( m_dyniLOD.GetValue() == slideoftSLF_OFF )
    {
      return;
    }

  // Set the NormalRenderLength based on the camera's Frustum
  vFrustum = m_pCSLIDECameraPath->GetCamera()->GetFrustumMax() -
             m_pCSLIDECameraPath->GetCamera()->GetFrustumMin();
  CSLIDEFace::SetNormalsRenderLength((FLOAT)(0.05 * MIN(vFrustum[X], vFrustum[Y])));


  // Get the Viewport Boundary
  vVPMax = vVPOrigin + vVPSize;

  // Create the matrix that goes from the canonical view volume
  // to the Viewport
  mGvp_proj.SetViewPort(vVPOrigin, vVPMax,
			m_vProjMin, m_vProjMax, 
			vFrustum[Y]/vFrustum[X]);

//////////////////////////////////////////////////////////////////////////////
// START: ASSIGNMENT 5: Fix this!!!
// 
  // The mGproj_vrc is the Camera's projection matrix
  mGproj_vrc.SetIdentity(); // ???
  mGvrc_proj.SetIdentity(); // ???

  // These lines will look backward but they are not
  // we need the inverse of the camera path
  mGvrc_obj.SetIdentity(); // ???
  mGobj_vrc.SetIdentity(); // ???
// 
// END  : ASSIGNMENT 5: Fix this!!!
//////////////////////////////////////////////////////////////////////////////

  if ( GetCrystalBallEnabled() )
    {
      mGvrc_obj = mGvrc_obj * m_mBall;
      mGobj_vrc = m_mBallInverse * mGobj_vrc;
    }

  m_pWorldRoot->Render(mGvp_proj, 
		       mGproj_vrc, mGvrc_proj,
		       mGvrc_obj, mGobj_vrc,
		       &m_boundClip, 
		       NULL,
		       CSLIDESurface::GetSLF_INHERIT(),
		       slideshtSLF_INHERIT, 
		       (SLIDELODType)m_dyniLOD.GetValue());

  // Render Clip Bound
  if ( GetRenderClipBound() )
    {
      gilColor3f(1.0, 1.0, 1.0);
      m_boundClip.Render(mGvp_proj);
    }
}
// 
// END  : ASSIGNMENT 5: Deal with the camera
//////////////////////////////////////////////////////////////////////////////

VOID CSLIDERender::WriteSLIDE(FILE *fpOutputFile)
{
  ASSERT( fpOutputFile != NULL );

  fprintf(fpOutputFile, "%s %s ", 
	  gcppcSLIDEReservedWords[sliderwtRender], 
	  m_pcViewportID);
  m_pCSLIDECameraPath->WriteSLIDE(fpOutputFile);
  fprintf(fpOutputFile, " %s\n", m_pcWorldRootID);

  if ( m_dyniLOD.GetValue() != slidelodtSLF_FULL )
    {
      fprintf(fpOutputFile, "  %s %s\n",
	      gcppcSLIDEReservedWords[sliderwtLOD], 
	      gcppcSLIDELODReservedWords[m_dyniLOD.GetValue()]);
    }

  fprintf(fpOutputFile, "%s\n", 
	  gcppcSLIDEReservedWords[sliderwtEndRender]);
}

//////////////////////////////////////////////////////////////////////
// CSLIDETclVariables Class
//

CSLIDETclVariables::CSLIDETclVariables()
{
  m_pcID = NULL;

  m_fWidth = 1.0;
  m_fHeight = 1.0;
  m_fMouseX = 0.0;
  m_fMouseY = 0.0;
}

CSLIDETclVariables::~CSLIDETclVariables()
{
  Uninit();
}

VOID CSLIDETclVariables::Init(CHAR *pcID)
{
  m_pcID = pcID;

  SetWidth(1.0);
  SetHeight(1.0);
  SetMouseX(0.0);
  SetMouseY(0.0);
}

VOID CSLIDETclVariables::Uninit()
{
}

BOOL CSLIDETclVariables::Initialized()
{
  return TRUE;
}

BOOL CSLIDETclVariables::Valid()
{
  return TRUE;
}

VOID CSLIDETclVariables::SetWidth(FLOAT fWidth)
{
  m_fWidth = fWidth;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_WIDTH");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_WIDTH", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fWidth);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetHeight(FLOAT fHeight)
{
  m_fHeight = fHeight;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_HEIGHT");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_HEIGHT", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fHeight);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseX(FLOAT fMouseX)
{
  m_fMouseX = fMouseX;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_X");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_X", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseX);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

VOID CSLIDETclVariables::SetMouseY(FLOAT fMouseY)
{
  m_fMouseY = fMouseY;

  if ( m_pcID == NULL )
    {
      sprintf(gpcBufferVariable, "SLF_MOUSE_Y");
    }
  else
    {
      sprintf(gpcBufferVariable, "SLF_%s_MOUSE_Y", m_pcID);
    }
  sprintf(gpcBufferValue, "%f", m_fMouseY);
  Tcl_SetVar(CSLIDEWorld::GetInterp(), gpcBufferVariable, 
	     gpcBufferValue, TCL_GLOBAL_ONLY);
}

